==> zoid-9.0.79.tgz: package/src/child/child.js <==
/* @flow */
/* eslint max-lines: 0 */

import { isSameDomain, matchDomain, getDomain, getOpener,
    getNthParentFromTop, getAncestor, getAllFramesInWindow,
    type CrossDomainWindowType, onCloseWindow, assertSameDomain } from 'cross-domain-utils/src';
import { markWindowKnown, deserializeMessage, type CrossDomainFunctionType } from 'post-robot/src';
import { ZalgoPromise } from 'zalgo-promise/src';
import { extend, onResize, elementReady, assertExists, noop } from 'belter/src';

import { getGlobal, tryGlobal } from '../lib';
import { CONTEXT, INITIAL_PROPS, WINDOW_REFERENCES } from '../constants';
import type { NormalizedComponentOptionsType, getSiblingsPropType } from '../component';
import type { PropsType, ChildPropsType } from '../component/props';
import type { WindowRef, PropRef, ParentExportsType } from '../parent';
import type { StringMatcherType } from '../types';

import { normalizeChildProps } from './props';
import { getChildPayload } from './window';

export type ChildExportsType<P> = {|
    updateProps : CrossDomainFunctionType<[ PropsType<P> ], void>,
    close : CrossDomainFunctionType<[], void>
|};

export type ChildHelpers<P, X> = {|
    uid : string,
    tag : string,
    close : () => ZalgoPromise<void>,
    focus : () => ZalgoPromise<void>,
    resize : ({| width : ?number, height : ?number |}) => ZalgoPromise<void>,
    onError : (mixed) => ZalgoPromise<void>,
    onProps : ((PropsType<P>) => void) => void,
    getParent : () => CrossDomainWindowType,
    getParentDomain : () => string,
    show : () => ZalgoPromise<void>,
    hide : () => ZalgoPromise<void>,
    export : (X) => ZalgoPromise<void>,
    getSiblings : getSiblingsPropType
|};

function getParentComponentWindow(ref : WindowRef) : CrossDomainWindowType {
    const { type } = ref;

    if (type === WINDOW_REFERENCES.OPENER) {
        return assertExists('opener', getOpener(window));

    } else if (type === WINDOW_REFERENCES.PARENT && typeof ref.distance === 'number') {
        return assertExists('parent', getNthParentFromTop(window, ref.distance));

    } else if (type === WINDOW_REFERENCES.GLOBAL && ref.uid && typeof ref.uid === 'string') {
        const { uid } = ref;
        const ancestor = getAncestor(window);

        if (!ancestor) {
            throw new Error(`Can not find ancestor window`);
        }

        for (const frame of getAllFramesInWindow(ancestor)) {
            if (isSameDomain(frame)) {
                const win = tryGlobal(frame, global => global.windows && global.windows[uid]);

                if (win) {
                    return win;
                }
            }
        }
    }

    throw new Error(`Unable to find ${ type } parent component window`);
}

function checkParentDomain(allowedParentDomains : StringMatcherType, domain : string) {
    if (!matchDomain(allowedParentDomains, domain)) {
        throw new Error(`Can not be rendered by domain: ${ domain }`);
    }
}

function focus() : ZalgoPromise<void> {
    return ZalgoPromise.try(() => {
        window.focus();
    });
}

function destroy() : ZalgoPromise<void> {
    return ZalgoPromise.try(() => {
        window.close();
    });
}

function getPropsByRef<P>(parentComponentWindow : CrossDomainWindowType, domain : string, propRef : PropRef) : (PropsType<P>) {
    let props;

    if (propRef.type === INITIAL_PROPS.RAW) {
        props = propRef.value;
    } else if (propRef.type === INITIAL_PROPS.UID) {
        if (!isSameDomain(parentComponentWindow)) {
            throw new Error(`Parent component window is on a different domain - expected ${ getDomain() } - can not retrieve props`);
        }

        const global = getGlobal(parentComponentWindow);
        props = assertExists('props', global && global.props[propRef.uid]);
    }

    if (!props) {
        throw new Error(`Could not find props`);
    }

    return deserializeMessage(parentComponentWindow, domain, props);
}

export type ChildComponent<P, X> = {|
    getProps : () => ChildPropsType<P, X>,
    init : () => ZalgoPromise<void>
|};

export function childComponent<P, X, C>(options : NormalizedComponentOptionsType<P, X, C>) : ChildComponent<P, X> {
    const { tag, propsDef, autoResize, allowedParentDomains } = options;

    const onPropHandlers = [];
    const childPayload = getChildPayload();
    let props : ChildPropsType<P, X>;
    const exportsPromise = new ZalgoPromise();

    if (!childPayload) {
        throw new Error(`No child payload found`);
    }

    if (childPayload.version !== __ZOID__.__VERSION__) {
        throw new Error(`Parent window has zoid version ${ childPayload.version }, child window has version ${ __ZOID__.__VERSION__ }`);
    }

    const { uid, parent: parentRef, parentDomain, exports: parentExports, context, props: propsRef } = childPayload;
    const parentComponentWindow = getParentComponentWindow(parentRef);
    const parent : ParentExportsType<P, X> = deserializeMessage(parentComponentWindow, parentDomain, parentExports);

    const { show, hide, close } = parent;

    const getParent = () => parentComponentWindow;
    const getParentDomain = () => parentDomain;
    
    const onProps = (handler : Function) => {
        onPropHandlers.push(handler);
    };

    const onError = (err : mixed) : ZalgoPromise<void> => {
        return ZalgoPromise.try(() => {
            if (parent && parent.onError) {
                return parent.onError(err);
            } else {
                throw err;
            }
        });
    };

    const resize = ({ width, height } : {| width : ?number, height : ?number |}) : ZalgoPromise<void> => {
        return parent.resize.fireAndForget({ width, height });
    };

    const xport = (xports : X) : ZalgoPromise<void> => {
        exportsPromise.resolve(xports);
        return parent.export(xports);
    };

    const getSiblings = ({ anyParent } = {}) => {
        const result = [];
        const currentParent = props.parent;

        if (typeof anyParent === 'undefined') {
            anyParent = !currentParent;
        }

        if (!anyParent && !currentParent) {
            throw new Error(`No parent found for ${ tag } child`);
        }

        for (const win of getAllFramesInWindow(window)) {
            if (!isSameDomain(win)) {
                continue;
            }

            const xprops : ChildPropsType<mixed, mixed> = assertSameDomain(win).xprops;

            if (!xprops || getParent() !== xprops.getParent()) {
                continue;
            }

            const winParent = xprops.parent;

            if (!anyParent && currentParent) {
                if (!winParent || winParent.uid !== currentParent.uid) {
                    continue;
                }
            }

            const xports = tryGlobal(win, global => global.exports);

            result.push({
                props:   xprops,
                exports: xports
            });
        }

        return result;
    };

    const getHelpers = () : ChildHelpers<P, X> => {
        return {
            tag, show, hide, close, focus, onError, resize, getSiblings,
            onProps, getParent, getParentDomain, uid, export: xport
        };
    };

    const watchForClose = () => {
        window.addEventListener('beforeunload', () => {
            parent.checkClose.fireAndForget();
        });

        window.addEventListener('unload', () => {
            parent.checkClose.fireAndForget();
        });

        onCloseWindow(parentComponentWindow, () => {
            destroy();
        });
    };

    const setProps = (newProps : PropsType<P>, origin : string, isUpdate : boolean = false) => {
        const helpers = getHelpers();
        const normalizedProps = normalizeChildProps(parentComponentWindow, propsDef, newProps, origin, helpers, isUpdate);

        if (props) {
            extend(props, normalizedProps);
        } else {
            props = normalizedProps;
        }

        for (const handler of onPropHandlers) {
            handler(props);
        }
    };
    
    const getAutoResize = () : ZalgoPromise<{| width : boolean, height : boolean, element : ?HTMLElement |}> => {
        const { width = false, height = false, element: elementRef = 'body' } = autoResize;
        return elementReady(elementRef).catch(noop).then(element => {
            return { width, height, element };
        });
    };

    const watchForResize = () : ?ZalgoPromise<void> => {
        return getAutoResize().then(({ width, height, element }) => {
            if (!element || (!width && !height) || context === CONTEXT.POPUP) {
                return;
            }

            onResize(element, ({ width: newWidth, height: newHeight }) => {
                resize({
                    width:  width ? newWidth : undefined,
                    height: height ? newHeight : undefined
                });
            }, { width, height });
        });
    };

    const updateProps = (newProps : (PropsType<P>)) : ZalgoPromise<void> => {
        return ZalgoPromise.try(() => setProps(newProps, parentDomain, true));
    };

    const init = () => {
        return ZalgoPromise.try(() => {
            getGlobal(window).exports = options.exports({
                getExports: () => exportsPromise
            });

            checkParentDomain(allowedParentDomains, parentDomain);
            markWindowKnown(parentComponentWindow);
            watchForClose();

            return parent.init({ updateProps, close: destroy });
    
        }).then(() => {
            return watchForResize();
    
        }).catch(err => {
            onError(err);
        });
    };

    const getProps = () => {
        if (props) {
            return props;
        }

        setProps(getPropsByRef(parentComponentWindow, parentDomain, propsRef), parentDomain);
        return props;
    };

    return {
        init,
        getProps
    };
}

==> zoid-9.0.82.tgz: package/src/child/child.js <==
/* @flow */
/* eslint max-lines: 0 */

import { isSameDomain, matchDomain, getAllFramesInWindow, type CrossDomainWindowType,
    onCloseWindow, assertSameDomain } from 'cross-domain-utils/src';
import { markWindowKnown, type CrossDomainFunctionType } from 'post-robot/src';
import { ZalgoPromise } from 'zalgo-promise/src';
import { extend, onResize, elementReady, noop } from 'belter/src';

import { getGlobal, tryGlobal, getInitialParentPayload, updateChildWindowNameWithRef } from '../lib';
import { CONTEXT } from '../constants';
import type { NormalizedComponentOptionsType, getSiblingsPropType } from '../component';
import type { PropsType, ChildPropsType } from '../component/props';
import type { StringMatcherType } from '../types';

import { normalizeChildProps } from './props';

export type ChildExportsType<P> = {|
    updateProps : CrossDomainFunctionType<[ PropsType<P> ], void>,
    close : CrossDomainFunctionType<[], void>
|};

export type ChildHelpers<P, X> = {|
    uid : string,
    tag : string,
    close : () => ZalgoPromise<void>,
    focus : () => ZalgoPromise<void>,
    resize : ({| width : ?number, height : ?number |}) => ZalgoPromise<void>,
    onError : (mixed) => ZalgoPromise<void>,
    onProps : ((PropsType<P>) => void) => void,
    getParent : () => CrossDomainWindowType,
    getParentDomain : () => string,
    show : () => ZalgoPromise<void>,
    hide : () => ZalgoPromise<void>,
    export : (X) => ZalgoPromise<void>,
    getSiblings : getSiblingsPropType
|};

function checkParentDomain(allowedParentDomains : StringMatcherType, domain : string) {
    if (!matchDomain(allowedParentDomains, domain)) {
        throw new Error(`Can not be rendered by domain: ${ domain }`);
    }
}

function focus() : ZalgoPromise<void> {
    return ZalgoPromise.try(() => {
        window.focus();
    });
}

function destroy() : ZalgoPromise<void> {
    return ZalgoPromise.try(() => {
        window.close();
    });
}

export type ChildComponent<P, X> = {|
    getProps : () => ChildPropsType<P, X>,
    init : () => ZalgoPromise<void>
|};

export function childComponent<P, X, C>(options : NormalizedComponentOptionsType<P, X, C>) : ChildComponent<P, X> {
    const { tag, propsDef, autoResize, allowedParentDomains } = options;

    const onPropHandlers = [];

    const { parent, payload } = getInitialParentPayload();
    const { win: parentComponentWindow, domain: parentDomain } = parent;

    let props : ChildPropsType<P, X>;
    const exportsPromise = new ZalgoPromise();

    const { version, uid, exports: parentExports, context, props: initialProps } = payload;

    if (version !== __ZOID__.__VERSION__) {
        throw new Error(`Parent window has zoid version ${ version }, child window has version ${ __ZOID__.__VERSION__ }`);
    }

    const { show, hide, close, onError, checkClose, export: parentExport, resize: parentResize, init: parentInit } = parentExports;

    const getParent = () => parentComponentWindow;
    const getParentDomain = () => parentDomain;
    
    const onProps = (handler : Function) => {
        onPropHandlers.push(handler);
    };

    const resize = ({ width, height } : {| width : ?number, height : ?number |}) : ZalgoPromise<void> => {
        return parentResize.fireAndForget({ width, height });
    };

    const xport = (xports : X) : ZalgoPromise<void> => {
        exportsPromise.resolve(xports);
        return parentExport(xports);
    };

    const getSiblings = ({ anyParent } = {}) => {
        const result = [];
        const currentParent = props.parent;

        if (typeof anyParent === 'undefined') {
            anyParent = !currentParent;
        }

        if (!anyParent && !currentParent) {
            throw new Error(`No parent found for ${ tag } child`);
        }

        for (const win of getAllFramesInWindow(window)) {
            if (!isSameDomain(win)) {
                continue;
            }

            const xprops : ChildPropsType<mixed, mixed> = assertSameDomain(win).xprops;

            if (!xprops || getParent() !== xprops.getParent()) {
                continue;
            }

            const winParent = xprops.parent;

            if (!anyParent && currentParent) {
                if (!winParent || winParent.uid !== currentParent.uid) {
                    continue;
                }
            }

            const xports = tryGlobal(win, global => global.exports);

            result.push({
                props:   xprops,
                exports: xports
            });
        }

        return result;
    };

    const getHelpers = () : ChildHelpers<P, X> => {
        return {
            tag, show, hide, close, focus, onError, resize, getSiblings,
            onProps, getParent, getParentDomain, uid, export: xport
        };
    };

    const watchForClose = () => {
        window.addEventListener('beforeunload', () => {
            checkClose.fireAndForget();
        });

        window.addEventListener('unload', () => {
            checkClose.fireAndForget();
        });

        onCloseWindow(parentComponentWindow, () => {
            destroy();
        });
    };

    const setProps = (newProps : PropsType<P>, origin : string, isUpdate : boolean = false) => {
        const helpers = getHelpers();
        const normalizedProps = normalizeChildProps(parentComponentWindow, propsDef, newProps, origin, helpers, isUpdate);

        if (props) {
            extend(props, normalizedProps);
        } else {
            props = normalizedProps;
        }

        for (const handler of onPropHandlers) {
            handler(props);
        }
    };
    
    const getAutoResize = () : ZalgoPromise<{| width : boolean, height : boolean, element : ?HTMLElement |}> => {
        const { width = false, height = false, element: elementRef = 'body' } = autoResize;
        return elementReady(elementRef).catch(noop).then(element => {
            return { width, height, element };
        });
    };

    const watchForResize = () : ?ZalgoPromise<void> => {
        return getAutoResize().then(({ width, height, element }) => {
            if (!element || (!width && !height) || context === CONTEXT.POPUP) {
                return;
            }

            onResize(element, ({ width: newWidth, height: newHeight }) => {
                resize({
                    width:  width ? newWidth : undefined,
                    height: height ? newHeight : undefined
                });
            }, { width, height });
        });
    };

    const updateProps = (newProps : (PropsType<P>)) : ZalgoPromise<void> => {
        return ZalgoPromise.try(() => setProps(newProps, parentDomain, true));
    };

    const init = () => {
        return ZalgoPromise.try(() => {
            if (isSameDomain(parentComponentWindow)) {
                updateChildWindowNameWithRef({
                    componentName: options.name,
                    parentComponentWindow
                });
            }

            getGlobal(window).exports = options.exports({
                getExports: () => exportsPromise
            });

            checkParentDomain(allowedParentDomains, parentDomain);
            markWindowKnown(parentComponentWindow);
            watchForClose();

            return parentInit({ updateProps, close: destroy });
    
        }).then(() => {
            return watchForResize();
    
        }).catch(err => {
            onError(err);
        });
    };

    const getProps = () => {
        if (props) {
            return props;
        } else {
            setProps(initialProps, parentDomain);
            return props;
        }
    };

    return {
        init,
        getProps
    };
}
